from fortuna.model.lenet import LeNet5
from fortuna.model.mlp import MLP
from fortuna.model.resnet import (ResNet18, ResNet34, ResNet50, ResNet101,
                                  ResNet152, ResNet200)
from fortuna.model.wideresnet import WideResNet28_10
